!----------------------------------------------------------
!         TALIst 
!----------------------------------------------------------
module AList

!===================================
!    TAR4list
!===================================
  Type :: TARlist
    character(32) VName
    logical, private :: IsDouble
    integer, private :: MaxSize, MaxSize0,NoSize
    real(4), pointer :: List4(:)
    real(4), pointer :: List8(:)
!   ---   -------    
    procedure(),NOPASS, pointer:: Error => NULL()  
  contains
    procedure,private :: reAlloc => TAR_ReAlloc
    procedure :: Create => TAR_Create
    procedure :: Free => TAR_Free
    procedure :: Get => TAR_Get
    procedure :: Count => TAR_GetCount
    procedure :: TAR_Set8
    procedure :: TAR_Set4
    GENERIC   :: Set => TAR_Set4, TAR_Set8
    
    procedure :: TAR_Getlist
    procedure :: TAR_GetONElist    
    procedure ::  TAR_GetMMMList
    GENERIC :: L => TAR_Getlist, TAR_GetOnelist, TAR_GetMMMList
    
    
  end type 

!===========
contains
!===========


!=====================================================
!         TARlist.Create
!=====================================================
 subroutine TAR_Create(this,VName, MaxSize0,IsDouble)
  class (TARlist) :: this
  logical, optional :: IsDouble
  logical L8
  character(*) VName
  integer  iRC
!---------------------------
!  
!---------------------------
  this%VName    = VName
  this%IsDouble = IsDouble
  L8 = .TRUE.
  if(PRESENT(IsDouble)) L8 = IsDouble
  if(IsDouble) then
     allocate(this%List8(MaxSize0), stat = iRC)
     nullify(this%List4)
     if(iRC /= 0) then
       if(associated(this%Error)) call this%Error(1,VName,MaxSize0)
       return
     endif
     this%List8 = 0.0
  else
     allocate(this%List4(MaxSize0), stat = iRC)
     nullify(this%List8)
     if(iRC /= 0) then
       if(associated(this%Error)) call this%Error(1,VName,MaxSize0)
       return
     endif
     this%List4 = 0.0
  endif
  this%MaxSize0 = MaxSize0
  this%MaxSize  = MaxSize0
  this%NoSize   = 0
  
 end subroutine 

!=====================================================
!         TARlist.Count
!=====================================================
 integer function TAR_GetCount(this)
  class (TARlist) :: this
!---------------------------
!    
!---------------------------
  TAR_GetCount = this%NoSize
 end function


!=====================================================
!         TARlist.Free
!=====================================================
 subroutine TAR_Free(this)
  class (TARlist) :: this
!---------------------------
!  
!---------------------------
  if(associated(this%List8)) deALLOCATE(this%List8)
  if(associated(this%List4)) deALLOCATE(this%List4)
 end subroutine
 
!=====================================================
!         TARlist.Get
!=====================================================
 real(8) function TAR_Get(this, IND) result(Ret) 
  class (TARlist) :: this
!---------------------------
!     IND
!---------------------------
  Ret = 0.0d0
  if(IND > 0 .AND. IND <= this%MaxSize) then
    if(this%IsDOUBLE) then
       Ret = this%List8(IND)
    else
       Ret = this%List4(IND)
    endif
    return
  else    
    if(associated(this%Error)) call this%Error(2,this%VName,IND)
  endif
 end function
 
!=====================================================
!         TARlist.Set4
!=====================================================
 subroutine TAR_Set4(this, IND,R4) 
  class (TARlist) :: this
  real(4) R4
!---------------------------
!   real(4)
!---------------------------
  if(IND < 1) return  
!-----   ----
  if(IND > this%MaxSize) then
     iDEL = IND - this%MaxSize 
     iAddSize = (iDEL/this%MaxSize0)* this%MaxSize0 + NINT(0.5*this%MaxSize0)       
     call this%Realloc(iAddSize,iRC)
     if(iRC /= 0) then
        if(associated(this%Error)) call this%Error(3,this%VName,this%MaxSize)
        return
     endif
  endif 
! ------  
  if(this%IsDOUBLE) then
       this%List8(IND) = R4
  else
       this%List4(IND) = R4
  endif
  this%NoSize = IND       
 end subroutine
 
!=====================================================
!         TARlist.Set8
!=====================================================
 subroutine TAR_Set8(this, IND,R8) 
  class (TARlist) :: this
  real(8) R8
!---------------------------
!   real(8)
!---------------------------
  if(IND < 1) return  
!-----   ----
  if(IND > this%MaxSize) then
     iDEL = IND - MaxSize 
     iAddSize = (iDEL/this%MaxSize0)* this%MaxSize0 + NINT(0.4*this%MaxSize0) 
     call this%Realloc(iAddSize,iRC)
     if(iRC /= 0) then
        if(associated(this%Error)) call this%Error(3,this%VName,this%MaxSize)
        return
     endif
  endif 
! ------  
  if(this%IsDOUBLE) then
       this%List8(IND) = R8
  else
       this%List4(IND) = R8
  endif
  this%NoSize = IND       
 end subroutine
 
!=====================================================
!         TARlist.Realloc
!=====================================================
 subroutine TAR_Realloc(this, AddSize,iRC) 
  class (TARlist) :: this
  integer addSize
  real, allocatable:: TMP4(:)
  real(8), allocatable:: TMP8(:)
!---------------------------
!  
!---------------------------
  iRC = 0 
  if(this%IsDOUBLE) then
      allocate(TMP8(this%MaxSize),stat=iRC)
      if(iRC /= 0) return
      TMP8 = this%List8
      deallocate(this%List8)
      allocate(this%List8(this%MaxSize + AddSize), stat=iRC)
      if(iRC /= 0) then
        deallocate(TMP8)
        return
      end if
      do m=1,this%MaxSize
        this%List8(m) = TMP8(m)
      end do
      do m=this%MaxSize+1, this%MaxSize+AddSize
        this%List8(m) = 0.0d0
      end do
      this%MaxSize = this%MaxSize+AddSize
      deallocate(TMP8)
  else
      allocate(TMP4(this%MaxSize),stat=iRC)
      if(iRC /= 0) return
      TMP4 = this%List4
      deallocate(this%List4)
      allocate(this%List4(this%MaxSize + AddSize), stat=iRC)
      if(iRC /= 0) then
        deallocate(TMP4)
        return
      end if
      do m=1,this%MaxSize
        this%List4(m) = TMP4(m)
      end do
      do m=this%MaxSize+1, this%MaxSize+AddSize
        this%List4(m) = 0.0d0
      end do
      this%MaxSize = this%MaxSize+AddSize
      deallocate(TMP4)
  endif
 end subroutine


!=========================================
!      
!=========================================
 subroutine CreateARlist(VName,RL,MaxSize,IsDOUBLE)
  Type (TARlist) RL
  character(*) VName
  logical, optional :: IsDouble
  logical LL
!-----
  LL = .FALSE.
  if(PRESENT(IsDouble)) LL = IsDouble  
  call RL%Create(VName,MaxSize,IsDOUBLE)
  RL%Error => ErrorZZ
 end subroutine

 subroutine ErrorZZ(iGO, VName, NoIT)
  character(*) VName
  
  if(iGO == 1) WRITE(*,*) 'Error: Allocate ERROR of "'//trim(VName)//'" , NoSize=',Noit
  if(iGO == 2) WRITE(*,*) 'Error: Index out list of "'//trim(VName)//'" , IND=',Noit
  if(iGO == 3) WRITE(*,*) 'Error: ReAllocate ERROR of "'//trim(VName)//'" , NoSize=',Noit
  STOP
 end subroutine 



!=====================================================
!         TARlist.GetList
!=====================================================
 function TAR_GetList(this) result(Ret)
  real(4),pointer :: Ret(:)  
  class (TARlist) :: this
!---------------------------
  Ret => this%List4(1:this%NoSize)
 end function

 function TAR_GetMMMList(this,I1,I2) result(Ret)
  real(4),pointer :: Ret(:)  
  class (TARlist) :: this
!---------------------------
  Ret => this%List4(I1:I2)
 end function

 real(4) function TAR_GetOneList(this,Ind) result(Ret)
  class (TARlist) :: this
!---------------------------
  Ret = this%List4(Ind)
 end function


end module

!==============================================
!      
!==============================================
program YESS
  USE AList
  Type (TARlist) A,B
  real, pointer:: ZZZ(:)
!------  
  character (40),parameter:: LFILE = 'output.txt'
  integer,parameter:: LU = 20
  logical flagFF
!---------
  INQUIRE(FILE=LFILE , EXIST = flagFF  )
  if(flagFF  ) then
             OPEN(LU, FILE = LFILE ,STATUS='old')
             CLOSE(LU ,STATUS='delete')
  end if
  OPEN(LU, FILE = LFILE ,STATUS='new')
!------
  call CreateARList('A',A,20,.FALSE.)
!------  
  
  call A%Set(2,12.5)
  write(LU,*) ' A(2)=', A%Get(2)

  call A%Set(124,66.6)
  
  write(LU,*) ' A(124)=', A%Get(124)
  
  write(LU,*) ' Count=', A%Count()
  write(LU,*) ' ****A(124)=', A%L(124)

  write(LU,*) ' ****A=', A%L()
  write(LU,*) ' ****A(124,125)=', A%L(124,125)
  ZZZ => A%L(124,125)
  write(LU,*) ' ****A(124,125)=', LBOUND(ZZZ),UBOUND(ZZZ)
  
  
call A%free
  

end program
